import copy
import inspect
import logging
import traceback
import typing

from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PackageLoader
from pydantic.dataclasses import dataclass
from starlette.datastructures import URL
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.trustedhost import TrustedHostMiddleware
from starlette.status import WS_1001_GOING_AWAY
from starlette.websockets import WebSocketClose

from .config import Config
from .http import responses
from .http.debug import ServerErrorMiddleware
from .http.exceptions import ExceptionMiddleware, HTTPException
from .http.request import Request
from .http.responses import RedirectResponse, Response
from .routing.routes import BaseRoute, NoMatchFound, Router
from .types import ASGIApp, Literal, Message, Receive, Scope, Send
from .websocket.request import WebSocket

logger = logging.getLogger(__name__)


def try_html(request: Request) -> typing.Optional[Response]:
    """
    try find html through TemplateResponse
    """
    try:
        return responses.TemplateResponse(
            request["path"] + ".html", {"request": request}
        )
    except LookupError:
        return None


class Lifespan:
    def __init__(
        self,
        on_startup: typing.List[typing.Callable] = None,
        on_shutdown: typing.List[typing.Callable] = None,
    ) -> None:
        self.on_startup = on_startup or []
        self.on_shutdown = on_shutdown or []

    def add_event_handler(
        self, event_type: Literal["startup", "shutdown"], func: typing.Callable
    ) -> None:
        if event_type == "startup":
            self.on_startup.append(func)
        elif event_type == "shutdown":
            self.on_shutdown.append(func)
        else:
            raise ValueError("event_type must be in ('startup', 'shutdown')")

    async def startup(self) -> None:
        """
        Run any `.on_startup` event handlers.
        """
        for handler in self.on_startup:
            result = handler()
            if inspect.isawaitable(result):
                await result

    async def shutdown(self) -> None:
        """
        Run any `.on_shutdown` event handlers.
        """
        for handler in self.on_shutdown:
            result = handler()
            if inspect.isawaitable(result):
                await result

    async def lifespan(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handle ASGI lifespan messages, which allows us to manage application
        startup and shutdown events.
        """
        message = await receive()
        assert message["type"] == "lifespan.startup"
        try:
            await self.startup()
        except BaseException:
            msg = traceback.format_exc()
            await send({"type": "lifespan.startup.failed", "message": msg})
            raise
        await send({"type": "lifespan.startup.complete"})

        message = await receive()
        assert message["type"] == "lifespan.shutdown"
        try:
            await self.shutdown()
        except BaseException:
            msg = traceback.format_exc()
            await send({"type": "lifespan.shutdown.failed", "message": msg})
            raise
        await send({"type": "lifespan.shutdown.complete"})

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "lifespan"
        await self.lifespan(scope, receive, send)


@dataclass
class FactoryClass:
    http: typing.Type[Request] = Request
    websocket: typing.Type[WebSocket] = WebSocket


class Index:
    def __init__(
        self,
        *,
        templates: typing.Iterable[str] = (),
        try_html: bool = True,
        on_startup: typing.List[typing.Callable] = [],
        on_shutdown: typing.List[typing.Callable] = [],
        routes: typing.List[BaseRoute] = [],
        middlewares: typing.List[Middleware] = [],
        exception_handlers: typing.Dict[
            typing.Union[int, typing.Type[Exception]], typing.Callable
        ] = {},
        factory_class: FactoryClass = FactoryClass(),
    ) -> None:
        self.factory_class = factory_class
        self.router = Router(routes)

        templates_loaders: typing.List[
            typing.Union[FileSystemLoader, PackageLoader]
        ] = []
        for template_path in templates:
            if ":" in template_path:  # package: "package:path"
                package_name, package_path = template_path.split(":", maxsplit=1)
                templates_loaders.append(PackageLoader(package_name, package_path))
            else:  # normal: "path"
                templates_loaders.append(FileSystemLoader(template_path))

        self.jinja_env = Environment(
            loader=ChoiceLoader(templates_loaders),
            enable_async=True,
        )
        self.try_html = try_html

        # Shallow copy list to prevent memory leak.
        self.lifespan = Lifespan(
            on_startup=copy.copy(on_startup), on_shutdown=copy.copy(on_shutdown)
        )

        self.user_middlewares = copy.copy(middlewares)
        self.exception_handlers = copy.copy(exception_handlers)

        self.asgiapp: ASGIApp = self.build_app()

    def rebuild_app(self) -> None:
        self.asgiapp = self.build_app()

    def build_app(self) -> ASGIApp:
        config = Config()
        error_handler = None
        exception_handlers = {}

        for key, value in self.exception_handlers.items():
            if key in (500, Exception):
                error_handler = value
            else:
                exception_handlers[key] = value

        middlewares = []

        if "*" in config.ALLOWED_HOSTS:
            middlewares.append(
                Middleware(TrustedHostMiddleware, allowed_hosts=config.ALLOWED_HOSTS)
            )

        middlewares.append(
            Middleware(ServerErrorMiddleware, handler=error_handler, debug=config.DEBUG)
        )

        if (
            config.CORS_ALLOW_ORIGIN_REGEX is not None
            or len(config.CORS_ALLOW_ORIGINS) > 0
        ):
            middlewares.append(
                Middleware(
                    CORSMiddleware,
                    allow_origins=config.CORS_ALLOW_ORIGINS,
                    allow_methods=config.CORS_ALLOW_METHODS,
                    allow_headers=config.CORS_ALLOW_HEADERS,
                    allow_credentials=config.CORS_ALLOW_CREDENTIALS,
                    allow_origin_regex=config.CORS_ALLOW_ORIGIN_REGEX,
                    expose_headers=config.CORS_EXPOSE_HEADERS,
                    max_age=config.CORS_MAX_AGE,
                )
            )

        middlewares += self.user_middlewares

        middlewares.append(Middleware(ExceptionMiddleware, handlers=exception_handlers))

        app = self.app

        for cls, options in reversed(middlewares):
            app = cls(app=app, **options)
        return app

    def add_middleware(self, middleware_class: type, **options: typing.Any) -> None:
        self.user_middlewares.insert(0, Middleware(middleware_class, **options))
        self.rebuild_app()

    def add_exception_handler(
        self,
        exc_class_or_status_code: typing.Union[int, typing.Type[Exception]],
        handler: typing.Callable,
    ) -> None:
        self.exception_handlers[exc_class_or_status_code] = handler
        self.rebuild_app()

    def exception_handler(
        self, exc_class_or_status_code: typing.Union[int, typing.Type[Exception]]
    ) -> typing.Callable:
        def decorator(func: typing.Callable) -> typing.Callable:
            self.add_exception_handler(exc_class_or_status_code, func)
            return func

        return decorator

    def on_startup(self, func: typing.Callable) -> typing.Callable:
        self.lifespan.add_event_handler("startup", func)
        return func

    def on_shutdown(self, func: typing.Callable) -> typing.Callable:
        self.lifespan.add_event_handler("shutdown", func)
        return func

    async def app(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        App without ASGI middleware.
        """
        if scope["type"] == "lifespan":
            return await self.lifespan(scope, receive, send)

        handler: typing.Optional[ASGIApp] = None

        try:
            path_params, handler = self.router.search(scope["type"], scope["path"])
            scope["path_params"] = path_params
        except NoMatchFound:
            if scope["type"] == "http" and self.try_html:
                # only html, no middleware/background tasks or other anything
                handler = try_html(self.factory_class.http(scope, receive, send))

        if handler is None:
            if scope["type"] == "http":
                raise HTTPException(404)
            handler = WebSocketClose(WS_1001_GOING_AWAY)

        return await handler(scope, receive, send)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self

        if (
            scope["type"] in ("http", "websocket")
            and Config().FORCE_SSL
            and scope["scheme"] in ("http", "ws")
        ):  # Force jump to https/wss
            url = URL(scope=scope)
            redirect_scheme = {"http": "https", "ws": "wss"}[url.scheme]
            netloc = url.hostname if url.port in (80, 443) else url.netloc
            url = url.replace(scheme=redirect_scheme, netloc=netloc)
            response = RedirectResponse(url, status_code=301)
            return await response(scope, receive, send)

        await self.asgiapp(scope, receive, send)


class Dispatcher:
    def __init__(
        self,
        default_app: ASGIApp,
        *apps: typing.Tuple[str, ASGIApp],
        handle404: ASGIApp = Response(b"", 404),
    ) -> None:
        self.default_app = default_app
        for prefix, app in apps:
            assert prefix.startswith("/"), "prefix must be start with '/'"
            assert not prefix.endswith("/"), "prefix can't end with '/'"
        self.apps = apps
        self.handle404 = handle404

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] in ("http", "websocket"):
            path = scope["path"]
            root_path = scope.get("root_path", "")
            has_received = False

            async def subreceive() -> Message:
                nonlocal has_received
                has_received = True
                return await receive()

            async def subsend(message: Message) -> None:
                if (
                    message["type"] == "http.response.start"
                    and message["status"] == 404
                ):
                    raise NoMatchFound()
                await send(message)

            # Call into a mounted app, if one exists.
            for path_prefix, app in filter(
                lambda item: path.startswith(item[0] + "/"), self.apps
            ):
                subscope = copy.copy(scope)
                subscope["path"] = path[len(path_prefix) :]
                subscope["root_path"] = root_path + path_prefix
                try:
                    return await app(subscope, subreceive, subsend)
                except NoMatchFound:
                    if has_received:  # has call received
                        return await self.handle404(scope, receive, send)

        await self.default_app(scope, receive, send)



import sys

from .applications import Dispatcher, Index
from .config import Config, here
from .utils import State

__all__ = ["Index", "Dispatcher", "Config", "g"]

# Current working directory first
sys.path.insert(0, here)

# global state
g = State()



import os
from copy import deepcopy
from typing import Any, Dict, List, Sequence, cast

from starlette.endpoints import Request, Response

from ..applications import Index
from ..http.responses import HTMLResponse, JSONResponse, YAMLResponse
from ..routing import HttpRoute
from ..types import Literal, TypedDict
from .schema import schema_parameter, schema_request_body, schema_response

Tag = TypedDict("Tag", {"description": str, "paths": Sequence[str]})


class OpenAPI:
    def __init__(
        self,
        title: str,
        description: str,
        version: str,
        *,
        tags: Dict[str, Tag] = {},
        template: str = "",
        media_type: Literal["yaml", "json"] = "yaml",
    ):
        """
        media_type: yaml or json
        """
        assert media_type in ("yaml", "json"), "media_type must in 'yaml' or 'json'"

        self.html_template = template
        self.media_type = media_type

        info = {"title": title, "description": description, "version": version}
        self.openapi = {
            "openapi": "3.0.0",
            "info": info,
            "paths": {},
            "tags": [
                {"name": tag_name, "description": tag_info.get("description", "")}
                for tag_name, tag_info in tags.items()
            ],
        }
        self.path2tag: Dict[str, List[str]] = {}
        for tag_name, tag_info in tags.items():
            for path in tag_info["paths"]:
                if path in self.path2tag:
                    self.path2tag[path].append(tag_name)
                else:
                    self.path2tag[path] = [tag_name]

    def _generate_paths(self, app: Index, definitions: dict) -> Dict[str, Any]:
        result = {}

        for path_format, endpoint in app.router.http_tree.iterator():
            path_docs = self._generate_path(endpoint, path_format, definitions)
            if not path_docs:
                continue
            result[path_format] = path_docs

        return result

    def _generate_path(self, view: Any, path: str, definitions: dict) -> Dict[str, Any]:
        result = {}
        if hasattr(view, "__methods__"):
            for method in view.__methods__:  # type: ignore
                if method == "OPTIONS":
                    continue
                method = method.lower()
                method_docs = self._generate_method(
                    getattr(view, method), path, definitions
                )
                if not method_docs:
                    continue
                result[method] = method_docs
        elif hasattr(view, "__method__"):
            method_docs = self._generate_method(view, path, definitions)
            if method_docs:
                result[view.__method__.lower()] = method_docs
        return result

    def _generate_method(
        self, func: object, path: str, definitions: dict
    ) -> Dict[str, Any]:
        result: Dict[str, Any] = {}

        doc = func.__doc__
        if isinstance(doc, str):
            doc = doc.strip()
            result.update(
                {
                    "summary": doc.splitlines()[0],
                    "description": "\n".join(doc.splitlines()[1:]).strip(),
                }
            )
        # generate params schmea
        params = getattr(func, "__params__", {})
        parameters = (
            schema_parameter(params.get("path"), "path")
            + schema_parameter(params.get("query"), "query")
            + schema_parameter(params.get("header"), "header")
            + schema_parameter(params.get("cookie"), "cookie")
        )
        if parameters:
            result["parameters"] = parameters
        # generate request body schmea
        request_body, _definitions = schema_request_body(params.get("body"))
        if request_body:
            result["requestBody"] = request_body
        definitions.update(_definitions)
        # generate responses schmea
        resps = getattr(func, "__resps__", {})
        responses: Dict[int, Any] = {}
        for status, content in resps.items():
            _ = responses[int(status)] = {"description": content["description"]}
            if content["model"] is None:
                continue
            _["content"], _definitions = schema_response(content["model"])
            definitions.update(_definitions)
        if responses:
            result["responses"] = responses
        # set path tags
        if result and path in self.path2tag:
            result["tags"] = self.path2tag[path]

        return result

    def create_docs(self, request: Request) -> dict:
        openapi: dict = deepcopy(self.openapi)
        definitions: Dict[str, Any] = {}
        openapi["servers"] = [
            {
                "url": f"{request.url.scheme}://{request.url.netloc}",
                "description": "Current server",
            }
        ]
        openapi["paths"] = self._generate_paths(
            cast(Index, request["app"]), definitions
        )
        openapi["definitions"] = definitions
        return openapi

    @property
    def routes(self) -> List[HttpRoute]:
        async def template(request: Request) -> Response:
            if self.html_template:
                return HTMLResponse(self.html_template)

            with open(
                os.path.join(
                    os.path.dirname(os.path.abspath(__file__)), "template.html"
                )
            ) as file:
                DEFAULT_TEMPLATE = file.read()
            return HTMLResponse(DEFAULT_TEMPLATE)

        async def docs(request: Request) -> Response:
            openapi = self.create_docs(request)
            media_type = request.query_params.get("type") or self.media_type

            if media_type == "json":
                return JSONResponse(openapi)
            return YAMLResponse(openapi)

        return [
            HttpRoute("/", template, name=None, method="get"),
            HttpRoute("/get", docs, name=None, method="get"),
        ]



import asyncio
import importlib
import inspect
import os
import threading
from functools import partial, update_wrapper
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple


class Singleton(type):
    def __init__(
        cls,
        name: str,
        bases: Tuple[type],
        namespace: Dict[str, Any],
    ) -> None:
        cls.instance = None
        super().__init__(name, bases, namespace)

    def __call__(cls, *args, **kwargs) -> Any:
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)
        return cls.instance


def import_module(name: str) -> Optional[ModuleType]:
    """
    try importlib.import_module, nothing to do when module not be found.
    """
    if os.path.exists(os.path.join(os.getcwd(), name + ".py")) or os.path.exists(
        os.path.join(os.getcwd(), name, "__init__.py")
    ):
        return importlib.import_module(name)
    return None  # nothing to do when module not be found


if TYPE_CHECKING:
    # https://github.com/python/mypy/issues/5107
    # for mypy check and IDE support
    cached_property = property
else:

    class cached_property:
        """
        A property that is only computed once per instance and then replaces
        itself with an ordinary attribute. Deleting the attribute resets the
        property.
        """

        def __init__(self, func: Callable) -> None:
            self.func = func
            update_wrapper(self, func)

        def __get__(self, obj: Any, cls: Any) -> Any:
            if obj is None:
                return self
            result = self.func(obj)
            if inspect.isawaitable(result):
                result = asyncio.ensure_future(result)
            value = obj.__dict__[self.func.__name__] = result
            return value


class superclass:
    """
    Call the method of the specified parent class. The usage is similar
    to `super`, but the difference from `super` is that `superclass`
    only looks for methods in the specified parent class.

    example:
        superclass(Class, obj).function(...)
    """

    def __init__(self, cls: type, instance: Any):
        if cls not in instance.__class__.mro():
            raise ValueError("`cls` must be in parent classes")

        self.__cls = cls
        self.__instance = instance

    def __getattr__(self, name: str) -> Callable[..., Any]:
        if not hasattr(self.__cls, name):
            raise AttributeError(name)

        func = getattr(self.__cls, name)
        return partial(func, self.__instance)


class State(dict):
    """
    An object that can be used to store arbitrary state.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.sync_lock = threading.Lock()
        self.async_lock = asyncio.Lock()

    def __enter__(self):
        self.sync_lock.acquire()
        return self

    def __exit__(self, exc_type, value, traceback):
        self.sync_lock.release()

    async def __aenter__(self):
        await self.async_lock.acquire()
        return self

    async def __aexit__(self, exc_type, value, traceback):
        self.async_lock.release()

    def __setattr__(self, name: Any, value: Any) -> None:
        self[name] = value

    def __getattr__(self, name: Any) -> Any:
        try:
            return self[name]
        except KeyError:
            message = "'{}' object has no attribute '{}'"
            raise AttributeError(message.format(self.__class__.__name__, name))

    def __delattr__(self, name: Any) -> None:
        del self[name]



import logging
import os
import signal
import subprocess
import sys
import time
from multiprocessing import cpu_count
from typing import List, Union

import click

from .__version__ import __version__
from .config import LOG_LEVELS, Config, here
from .utils import import_module

config = Config()


def execute(command: Union[List[str], str]) -> int:
    if isinstance(command, str):
        command = [command]

    click.echo("Execute command: ", nl=False)
    click.secho(" ".join(command), fg="green")

    process = subprocess.Popen(" ".join(command), shell=True)

    def sigterm_handler(signo, frame):
        process.terminate()
        process.wait()

    signal.signal(signal.SIGINT, sigterm_handler)
    signal.signal(signal.SIGTERM, sigterm_handler)

    while process.poll() is None:
        time.sleep(1)

    return process.returncode


@click.group(help=f"Index.py {__version__}")
def index_cli():
    # set index logger level
    indexpy_logger = logging.getLogger("indexpy")
    indexpy_logger.setLevel(LOG_LEVELS[config.LOG_LEVEL])


@index_cli.command(help="use uvicorn to run Index.py")
@click.argument("application", default=lambda: config.APP)
def serve(application):
    import uvicorn

    uvicorn.run(
        application,
        host=config.HOST,
        port=config.PORT,
        log_level=config.LOG_LEVEL,
        interface="asgi3",
        lifespan="on",
        reload=config.AUTORELOAD,
    )


@click.group(help="use gunicorn to run Index.py")
def gunicorn():
    pass


@gunicorn.command()
@click.option("--workers", "-w", default=cpu_count())
@click.option("--worker-class", "-k", default="uvicorn.workers.UvicornWorker")
@click.option("--daemon", "-d", default=False, is_flag=True)
@click.option(
    "--configuration",
    "-c",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
)
@click.argument("application", default=lambda: config.APP)
def start(workers, worker_class, daemon, configuration, application):
    from gunicorn.app.wsgiapp import run

    command = (
        "gunicorn"
        + f" -k {worker_class}"
        + f" --bind {config.HOST}:{config.PORT}"
        + f" --chdir {here}"
        + f" --workers {workers}"
        + f" --pid {os.path.join(here, '.pid')}"
        + f" --log-level {config.LOG_LEVEL}"
    )
    args = command.split(" ")
    if daemon:
        args.extend("-D --log-file log.index".split(" "))
    if config.AUTORELOAD:
        args.append("--reload")
    if configuration:
        args.append("-c")
        args.append(configuration.strip())
    args.append(application)

    sys.argv = args
    run()


@gunicorn.command("")
def stop():
    execute(["kill -TERM", "`cat .pid`"])


@gunicorn.command()
def reload():
    execute(["kill -HUP", "`cat .pid`"])


index_cli.add_command(gunicorn, "gunicorn")

import_module("commands")



import copy
import importlib
import os
import typing
from dataclasses import InitVar, asdict, dataclass
from functools import update_wrapper, wraps
from pathlib import Path

from indexpy.concurrency import complicating
from indexpy.http.responses import convert
from indexpy.http.view import only_allow
from indexpy.types import ASGIApp, Literal, Receive, Scope, Send
from indexpy.utils import superclass

from .convertors import Convertor, compile_path
from .tree import RadixTree


def request_response(view: typing.Any) -> ASGIApp:
    @wraps(view)
    async def _(scope: Scope, receive: Receive, send: Send) -> None:
        current_app = scope["app"]
        request = current_app.factory_class.http(scope, receive, send)
        response = convert(await view(request))
        await response(scope, receive, send)

    return _


def websocket_session(view: typing.Any) -> ASGIApp:
    @wraps(view)
    async def _(scope: Scope, receive: Receive, send: Send) -> None:
        current_app = scope["app"]
        websocket = current_app.factory_class.websocket(scope, receive, send)
        await view(websocket)(scope, receive, send)

    return _


def subpath_asgi(path_prefix: str, asgi: ASGIApp) -> ASGIApp:
    assert path_prefix.startswith("/"), "path_prefix must be start with '/'"
    assert not path_prefix.endswith("/"), "path_prefix can't end with '/'"

    @wraps(asgi)
    async def _(scope: Scope, receive: Receive, send: Send) -> None:
        path = scope["path"]
        root_path = scope.get("root_path", "")

        subscope = copy.copy(scope)
        subscope["path"] = path[len(path_prefix) :]
        subscope["root_path"] = root_path + path_prefix
        await asgi(subscope, receive, send)

    return _


class NoMatchFound(Exception):
    """
    Raised by `.search(path)` if no matching route exists.
    """


class NoRouteFound(Exception):
    """
    Raised by `.url_for(name, **path_params)` if no matching route exists.
    """


@dataclass
class BaseRoute:
    path: str
    endpoint: typing.Any
    name: typing.Optional[str] = ""

    def extend_middlewares(self, routes: typing.List["BaseRoute"]) -> None:
        raise NotImplementedError()

    def __post_init__(self) -> None:
        if not self.path.startswith("/"):
            raise ValueError("Route path must start with '/'")
        if self.name == "":
            self.name = self.endpoint.__name__


@dataclass
class HttpRoute(BaseRoute):
    method: InitVar[str] = ""

    def __post_init__(self, method: str) -> None:  # type: ignore
        super().__post_init__()

        self.endpoint = complicating(self.endpoint)

        if not (
            hasattr(self.endpoint, "__methods__")
            or hasattr(self.endpoint, "__method__")
        ):
            if method == "":
                raise ValueError("View function must be marked with method")
            self.endpoint = only_allow(method, self.endpoint)
        else:
            if hasattr(self.endpoint, "__method__") and (
                method.upper() not in (getattr(self.endpoint, "__method__"), "")
            ):
                raise ValueError("View function has been marked with method")
            if hasattr(self.endpoint, "__methods__") and method != "":
                raise ValueError("View class can't be marked with method")

    def extend_middlewares(self, routes: typing.List[BaseRoute]) -> None:
        if getattr(routes, "_http_middlewares", None):
            endpoint = self.endpoint
            for middleware in getattr(routes, "_http_middlewares"):
                endpoint = middleware(endpoint)
            self.endpoint = update_wrapper(endpoint, self.endpoint)


@dataclass
class SocketRoute(BaseRoute):
    def extend_middlewares(self, routes: typing.List[BaseRoute]) -> None:
        if getattr(routes, "_socket_middlewares", None):
            endpoint = self.endpoint
            for middleware in getattr(routes, "_socket_middlewares"):
                endpoint = middleware(endpoint)
            self.endpoint = update_wrapper(endpoint, self.endpoint)


@dataclass
class ASGIRoute(BaseRoute):
    # This is mypy error
    type: typing.Container[Literal["http", "websocket"]] = ("http", "websocket")  # type: ignore
    root_path: InitVar[str] = ""

    def __post_init__(self, root_path: str) -> None:  # type: ignore
        super().__post_init__()
        if root_path:
            self.endpoint = subpath_asgi(root_path, self.endpoint)

    def extend_middlewares(self, routes: typing.List[BaseRoute]) -> None:
        if hasattr(routes, "_asgi_middlewares"):
            endpoint = self.endpoint
            for middleware in getattr(routes, "_asgi_middlewares"):
                endpoint = middleware(endpoint)
            self.endpoint = update_wrapper(endpoint, self.endpoint)


T = typing.TypeVar("T")


class RouteRegisterMixin:
    __slots__ = ()

    def append(self, route: BaseRoute) -> None:
        raise NotImplementedError()

    def extend(self, routes: typing.List[BaseRoute]) -> None:
        for route in routes:  # type: BaseRoute
            if isinstance(routes, Routes):
                route.extend_middlewares(routes)

            if isinstance(routes, SubRoutes):
                data = asdict(route)
                data["path"] = routes.prefix + data["path"]
                route = route.__class__(**data)

            if getattr(routes, "namespace", None) and route.name:
                route.name = getattr(routes, "namespace") + ":" + route.name

            self.append(route)

    def http(
        self, path: str, *, name: str = "", method: str = ""
    ) -> typing.Callable[[T], T]:
        """
        shortcut for `self.append(HttpRoute(path, endpoint, name, method))`

        example:
            @routes.http("/path", name="endpoint-name")
            class Endpoint(HTTPView): ...
        """

        def register(endpoint: T) -> T:
            self.append(HttpRoute(path, endpoint, name, method))
            return endpoint

        return register

    def websocket(self, path: str, *, name: str = "") -> typing.Callable[[T], T]:
        """
        shortcut for `self.append(SocketRoute(path, endpoint, name))`

        example:
            @routes.websocket("/path", name="endpoint-name")
            async def endpoint(websocket): ...
        """

        def register(endpoint: T) -> T:
            self.append(SocketRoute(path, endpoint, name))
            return endpoint

        return register

    def asgi(
        self,
        path: str,
        *,
        name: str = "",
        type: typing.Container[Literal["http", "websocket"]] = ("http", "websocket"),  # type: ignore
        root_path: str = "",
    ) -> typing.Callable[[T], T]:
        """
        shortcut for `self.append(ASGIRoute(path, endpoint, name, type))`

        example:
            @routes.asgi("/path", name="endpoint-name")
            class Endpoint(HTTPView): ...
        """

        def register(endpoint: T) -> T:
            self.append(ASGIRoute(path, endpoint, name, type, root_path))
            return endpoint

        return register


class Routes(typing.List[BaseRoute], RouteRegisterMixin):
    def __init__(
        self,
        *iterable: typing.Union["BaseRoute", "Routes", "FileRoutes"],
        namespace: str = "",
        http_middlewares: typing.List[typing.Any] = [],
        socket_middlewares: typing.List[typing.Any] = [],
        asgi_middlewares: typing.List[typing.Any] = [],
    ) -> None:
        super().__init__()
        self.namespace = namespace
        self._http_middlewares = copy.copy(http_middlewares)
        self._socket_middlewares = copy.copy(socket_middlewares)
        self._asgi_middlewares = copy.copy(asgi_middlewares)
        for route in iterable:
            if not isinstance(route, typing.List):
                self.append(route)
            else:
                self.extend(route)

    def extend(self, routes: typing.List[BaseRoute]) -> None:  # type: ignore
        """
        Extend routes in routes

        example:
            routes.extend(Routes(...))
        or
            routes.extend([...])
        """
        return superclass(RouteRegisterMixin, self).extend(routes)

    def http_middleware(self, middleware: T) -> T:
        """
        append middleware in routes

        example:
            @routes.http_middleware
            def middleware(endpoint):
                async def wrapper(request):
                    response = await endpoint(request)
                    return response
                return wrapper
        """
        self._http_middlewares.append(middleware)
        return middleware

    def socket_middleware(self, middleware: T) -> T:
        """
        append middleware in routes

        example:
            @routes.socket_middleware
            def middleware(endpoint):
                async def wrapper(websocket):
                    await endpoint(websocket)
                return wrapper
        """
        self._socket_middlewares.append(middleware)
        return middleware

    def asgi_middleware(self, middleware: T) -> T:
        """
        append middleware in routes

        example:
            @routes.asgi_middleware
            def middleware(endpoint):
                async def wrapper(scope, receive, send):
                    await endpoint(scope, receive, send)
                return wrapper
        """
        self._asgi_middlewares.append(middleware)
        return middleware


class SubRoutes(Routes):
    def __init__(
        self,
        prefix: str,
        routes: typing.List[BaseRoute] = [],
        *,
        namespace: str = "",
        http_middlewares: typing.List[typing.Any] = [],
        socket_middlewares: typing.List[typing.Any] = [],
        asgi_middlewares: typing.List[typing.Any] = [],
    ) -> None:
        if not prefix.startswith("/") or prefix.endswith("/"):
            raise ValueError("Mount prefix cannot end with '/' and must start with '/'")
        self.prefix = prefix
        super().__init__(
            *routes,
            namespace=namespace,
            http_middlewares=http_middlewares,
            socket_middlewares=socket_middlewares,
            asgi_middlewares=asgi_middlewares,
        )


class FileRoutes(typing.List[BaseRoute]):
    def __init__(
        self,
        module_name: str,
        *,
        namespace: str = "",
        allow_underline: bool = False,
        suffix: str = "",
    ) -> None:
        dirpath = Path(
            os.path.abspath(importlib.import_module(module_name).__file__)
        ).parent
        assert dirpath.name == module_name

        for pypath in dirpath.glob("**/*.py"):
            relpath = str(pypath.relative_to(dirpath)).replace("\\", "/")[:-3]

            path_list = relpath.split("/")
            path_list.insert(0, module_name)

            url_path = "/" + relpath
            if not allow_underline:
                url_path = url_path.replace("_", "-")
            if url_path.endswith("/index"):
                url_path = url_path[: -len("index")]
            else:
                url_path = url_path + suffix

            module = importlib.import_module(".".join(path_list))
            url_name = getattr(module, "name", None)
            get_response = getattr(module, "HTTP", None)
            serve_socket = getattr(module, "Socket", None)
            if get_response:
                for deep in range(len(path_list), 0, -1):
                    _module = importlib.import_module(".".join(path_list[:deep]))
                    if not hasattr(_module, "HTTPMiddleware"):
                        continue
                    get_response = getattr(_module, "HTTPMiddleware")(get_response)
                self.append(HttpRoute(url_path, get_response, url_name))
            if serve_socket:
                for deep in range(len(path_list), 0, -1):
                    _module = importlib.import_module(".".join(path_list[:deep]))
                    if not hasattr(_module, "SocketMiddleware"):
                        continue
                    serve_socket = getattr(_module, "SocketMiddleware")(serve_socket)
                self.append(SocketRoute(url_path, serve_socket, url_name))

        self.namespace = namespace


class Router(RouteRegisterMixin):
    def __init__(self, routes: typing.List[BaseRoute] = list()) -> None:
        self.http_tree = RadixTree()
        self.websocket_tree = RadixTree()

        self.http_routes: typing.Dict[
            str, typing.Tuple[str, typing.Dict[str, Convertor], ASGIApp]
        ] = {}
        self.websocket_routes: typing.Dict[
            str, typing.Tuple[str, typing.Dict[str, Convertor], ASGIApp]
        ] = {}

        self.extend(routes)

    def _append(
        self,
        path: str,
        endpoint: ASGIApp,
        name: typing.Optional[str],
        radix_tree: RadixTree,
        routes: typing.Dict,
    ) -> None:
        if name in routes:
            raise ValueError(f"Duplicate route name: {name}")

        radix_tree.append(path, endpoint)
        path_format, path_convertors = compile_path(path)

        if name:  # name not in ("", None)
            routes[name] = (path_format, path_convertors, endpoint)

    def append(self, route: BaseRoute) -> None:
        if isinstance(route, HttpRoute):
            self._append(
                route.path,
                request_response(route.endpoint),
                route.name,
                self.http_tree,
                self.http_routes,
            )
        elif isinstance(route, SocketRoute):
            self._append(
                route.path,
                websocket_session(route.endpoint),
                route.name,
                self.websocket_tree,
                self.websocket_routes,
            )
        elif isinstance(route, ASGIRoute):
            if "http" in route.type:
                self._append(
                    route.path,
                    route.endpoint,
                    route.name,
                    self.http_tree,
                    self.http_routes,
                )
            if "websocket" in route.type:
                self._append(
                    route.path,
                    route.endpoint,
                    route.name,
                    self.websocket_tree,
                    self.websocket_routes,
                )
        else:
            raise TypeError(
                "Need type: `ASGIRoute`, `HttpRoute` or `SocketRoute`,"
                + f" but got type: {type(route)}"
            )

    def search(
        self, protocol: Literal["http", "websocket"], path: str
    ) -> typing.Tuple[typing.Dict[str, typing.Any], ASGIApp]:
        if protocol == "http":
            radix_tree = self.http_tree
        elif protocol == "websocket":
            radix_tree = self.websocket_tree
        else:
            raise ValueError("`protocol` must be in ('http', 'websocket')")

        params, endpoint = radix_tree.search(path)

        if params is None or endpoint is None:
            raise NoMatchFound(path)

        return params, endpoint

    def extend(self, routes: typing.List[BaseRoute]) -> None:
        """
        Add routes in router

        example:
            router.extend(Routes(...))
        or
            router.extend([...])
        """
        return superclass(RouteRegisterMixin, self).extend(routes)

    def url_for(
        self,
        name: str,
        path_params: typing.Dict[str, typing.Any] = {},
        protocol: Literal["http", "websocket"] = "http",
    ) -> str:
        if protocol == "http":
            routes = self.http_routes
        elif protocol == "websocket":
            routes = self.websocket_routes
        else:
            raise ValueError("`protocol` must be in ('http', 'websocket')")

        if name not in routes:
            raise NoRouteFound(f"No route with name '{name}' exists")

        path_format, path_convertors, _ = routes[name]

        return path_format.format_map(
            {
                name: path_convertors[name].to_string(value)
                for name, value in path_params.items()
            }
        )



import math
import re
import typing
import uuid
from decimal import Decimal


class Convertor:
    regex = ""

    def convert(self, value: str) -> typing.Any:
        raise NotImplementedError()

    def to_string(self, value: typing.Any) -> str:
        raise NotImplementedError()


class StringConvertor(Convertor):
    regex = "[^/]+"

    def convert(self, value: str) -> str:
        return value

    def to_string(self, value: typing.Any) -> str:
        value = str(value)
        if not value:
            raise ValueError("Must not be empty")
        if "/" in value:
            raise ValueError("May not contain path separators")
        return value


class PathConvertor(Convertor):
    regex = ".*"

    def convert(self, value: str) -> str:
        return str(value)

    def to_string(self, value: typing.Any) -> str:
        return str(value)


class IntegerConvertor(Convertor):
    regex = "[0-9]+"

    def convert(self, value: str) -> int:
        return int(value)

    def to_string(self, value: typing.Any) -> str:
        value = int(value)
        if value < 0:
            raise ValueError("Negative integers are not supported")
        return str(value)


class DecimalConvertor(Convertor):
    regex = "[0-9]+(.[0-9]+)?"

    def convert(self, value: str) -> Decimal:
        return Decimal(value)

    def to_string(self, value: typing.Any) -> str:
        value = Decimal(value)
        if value < Decimal("0.0"):
            raise ValueError("Negative decimal are not supported")
        if math.isnan(value):
            raise ValueError("NaN values are not supported")
        if math.isinf(value):
            raise ValueError("Infinite values are not supported")
        return str(value).rstrip("0").rstrip(".")


class UUIDConvertor(Convertor):
    regex = "[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}"

    def convert(self, value: str) -> uuid.UUID:
        return uuid.UUID(value)

    def to_string(self, value: uuid.UUID) -> str:
        return str(value)


CONVERTOR_TYPES = {
    "str": StringConvertor(),
    "path": PathConvertor(),
    "int": IntegerConvertor(),
    "decimal": DecimalConvertor(),
    "uuid": UUIDConvertor(),
}

# Match parameters in URL paths, eg. '{param}', and '{param:int}'
PARAM_REGEX = re.compile(r"{([a-zA-Z_][a-zA-Z0-9_]*)(:[a-zA-Z_][a-zA-Z0-9_]*)?}")


def is_compliant(path: str) -> bool:
    """
    Whether the "{...}" are closed
    """
    unclosed_count = 0
    for c in path:
        if c == "{":
            unclosed_count += 1
        elif c == "}":
            unclosed_count -= 1
    return unclosed_count == 0


def compile_path(path: str) -> typing.Tuple[str, typing.Dict[str, Convertor]]:
    """
    Given a path string, like: "/{username:str}", return a two-tuple
    of (format, {param_name:convertor}).

    format:     "/{username}"
    convertors: {"username": StringConvertor()}
    """
    if not is_compliant(path):
        raise ValueError(f"There are unclosed braces: {path}")

    path_format = ""

    idx = 0
    param_convertors = {}
    for match in PARAM_REGEX.finditer(path):
        param_name, convertor_type = match.groups("str")
        convertor_type = convertor_type.lstrip(":")
        if convertor_type not in CONVERTOR_TYPES:
            raise ValueError(f"Unknown path convertor '{convertor_type}'")
        convertor = CONVERTOR_TYPES[convertor_type]

        path_format += path[idx : match.start()]
        path_format += "{%s}" % param_name

        param_convertors[param_name] = convertor

        idx = match.end()

    path_format += path[idx:]

    return path_format, param_convertors



import typing

from .request import WebSocket


class MiddlewareMixin:

    mounts: typing.Sequence[typing.Callable] = ()

    def __init__(self, get_response: typing.Callable) -> None:
        self.get_response = self.mount_middleware(get_response)

    def mount_middleware(self, get_response: typing.Callable) -> typing.Callable:
        for middleware in reversed(self.mounts):
            get_response = middleware(get_response)
        return get_response

    async def __call__(self, websocket: WebSocket) -> None:
        await self.before_accept(websocket)
        await self.get_response(websocket)
        await self.after_close(websocket)

    async def before_accept(self, websocket: WebSocket) -> None:
        """
        Called before calling websocket handler
        """

    async def after_close(self, websocket: WebSocket) -> None:
        """
        Called after the websocket handler has finished processing
        """



from copy import deepcopy
from typing import Any, Dict, List, Tuple, Optional, Union

import yaml
from pydantic import BaseModel

from ..types import Literal
from .types import File


def schema_parameter(
    m: Optional[BaseModel],
    position: Literal["path", "query", "header", "cookie"],
) -> List[Dict[str, Any]]:
    result = []

    if m is not None:
        _schemas = deepcopy(m.schema())
        properties: Dict[str, Any] = _schemas["properties"]
        required = _schemas.get("required", ())

        for name, schema in properties.items():  # type: str, Dict[str, str]
            result.append(
                {
                    "in": position,
                    "name": name,
                    "description": schema.pop("description", ""),
                    "required": name in required,  # type: ignore
                    "schema": schema,
                }
            )
    return result


def schema_request_body(body: BaseModel = None) -> Tuple[Optional[Dict], Dict]:
    if body is None:
        return None, {}

    _schema: Dict = deepcopy(body.schema())
    definitions = _schema.pop("definitions", {})

    for field in body.__fields__.values():
        if issubclass(field.type_, File):
            return {
                "required": True,
                "content": {"multipart/form-data": {"schema": _schema}},
            }, definitions

    return {
        "required": True,
        "content": {"application/json": {"schema": _schema}},
    }, definitions


def schema_response(model: Union[BaseModel, str]) -> Tuple[Dict, Dict]:
    if isinstance(model, str):
        return yaml.safe_load(model.strip()), {}
    schema = deepcopy(model.schema())
    definitions = schema.pop("definitions", {})
    return {"application/json": {"schema": schema}}, definitions



import sys
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Callable,
    Iterable,
    MutableMapping,
    Optional,
    Tuple,
    Type,
)

if sys.version_info[:2] < (3, 8):
    from typing_extensions import Final, Literal, TypedDict, final
else:  # pragma: no cover
    from typing import Final, Literal, TypedDict, final

__all__ = [
    "Scope",
    "Message",
    "Receive",
    "Send",
    "ASGIApp",
    "ExcInfo",
    "Environ",
    "StartResponse",
    "WSGIApp",
] + [
    # built-in types
    "TypedDict",
    "Literal",
    "Final",
    "final",
]

# ASGI
Scope = MutableMapping[str, Any]

Message = MutableMapping[str, Any]

Receive = Callable[[], Awaitable[Message]]

Send = Callable[[Message], Awaitable[None]]

ASGIApp = Callable[[Scope, Receive, Send], Awaitable[None]]

# WSGI: view PEP3333
ExcInfo = Tuple[Type[BaseException], BaseException, Optional[TracebackType]]

Environ = MutableMapping[str, Any]

StartResponse = Callable[[str, Iterable[Tuple[str, str]], Optional[ExcInfo]], None]

WSGIApp = Callable[[Environ, StartResponse], Iterable[bytes]]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['indexpy',
 'indexpy.http',
 'indexpy.openapi',
 'indexpy.routing',
 'indexpy.websocket']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'jinja2>=2.10.3,<3.0.0',
 'pydantic>=1.6,<2.0',
 'python-multipart>=0.0.5,<0.0.6',
 'pyyaml>=5.3,<6.0',
 'starlette>=0.13.6,<0.14.0']

extras_require = \
{':python_version < "3.8"': ['typing-extensions>=3.7.4,<4.0.0'],
 'full': ['gunicorn>=20.0.4,<21.0.0',
          'requests>=2.24.0,<3.0.0',
          'uvicorn>=0.12.1,<0.13.0'],
 'serve': ['gunicorn>=20.0.4,<21.0.0', 'uvicorn>=0.12.1,<0.13.0'],
 'test': ['requests>=2.24.0,<3.0.0']}

entry_points = \
{'console_scripts': ['index-cli = indexpy.cli:index_cli']}

setup_kwargs = {
    'name': 'index.py',
    'version': '0.13.1',
    'description': 'An easy-to-use asynchronous web framework based on ASGI.',
    'long_description': '<div align="center">\n\n<img style="max-width:60%;" src="https://raw.githubusercontent.com/abersheeran/index.py/master/docs/img/index-py.png" />\n\n<p>\n中文\n|\n<a href="https://github.com/abersheeran/index.py/tree/master/README-en.md">English</a>\n</p>\n\n<p>\n<a href="https://github.com/abersheeran/index.py/actions?query=workflow%3ATest">\n<img src="https://github.com/abersheeran/index.py/workflows/Test/badge.svg" alt="Github Action Test" />\n</a>\n\n<a href="https://github.com/abersheeran/index.py/actions?query=workflow%3A%22Build+setup.py%22">\n<img src="https://github.com/abersheeran/index.py/workflows/Build%20setup.py/badge.svg" alt="Build setup.py" />\n</a>\n</p>\n\n<p>\n<a href="https://github.com/abersheeran/index.py/actions?query=workflow%3A%22Publish+PyPi%22">\n<img src="https://github.com/abersheeran/index.py/workflows/Publish%20PyPi/badge.svg" alt="Publish PyPi" />\n</a>\n\n<a href="https://pypi.org/project/index.py/">\n<img src="https://img.shields.io/pypi/v/index.py" alt="PyPI" />\n</a>\n\n<a href="https://pepy.tech/project/index-py/week">\n<img src="https://pepy.tech/badge/index-py/week" alt="Week Downloads">\n</a>\n</p>\n\n<p>\n<img src="https://img.shields.io/pypi/pyversions/index.py" alt="PyPI - Python Version" />\n</p>\n\n一个基于 Radix Tree 的高性能 web 框架。\n\n<a href="https://index-py.abersheeran.com">Index.py 文档</a>\n\n</div>\n\n---\n\nIndex.py 实现了 [ASGI3](http://asgi.readthedocs.io/en/latest/) 接口，并使用 Radix Tree 进行路由查找。是最快的 Python web 框架之一。一切特性都服务于快速开发高性能的 Web 服务。\n\n- 灵活且高效的路由系统 (基于 Radix Tree)\n- 自动解析请求 & 生成文档 (基于 [pydantic](https://pydantic-docs.helpmanual.io/))\n- 可视化 API 接口 (基于 ReDoc, 针对中文字体优化)\n- 自带一键部署命令 (基于 uvicorn 与 gunicorn)\n- 挂载 ASGI/WSGI 应用\n- 进程内后台任务 (基于 [asyncio](https://docs.python.org/3/library/asyncio.html))\n- 可使用任何可用的 ASGI 生态\n\n## Install\n\n```bash\npip install -U index.py\n```\n\n或者直接从 Github 上安装最新版本（不稳定）\n\n```bash\npip install -U git+https://github.com/abersheeran/index.py@setup.py\n```\n\n中国大陆内的用户可从 Coding 上的镜像仓库拉取\n\n```bash\npip install -U git+https://e.coding.net/aber/github/index.py.git@setup.py\n```\n\n## Quick start\n\n向一个 `.py` 文件写入如下代码并直接执行它，访问 `http://127.0.0.1:4190`。\n\n```python\nfrom indexpy import Index\n\n\napp = Index()\n\n\n@app.router.http("/", method="get")\nasync def homepage(request):\n    return "hello, index.py"\n\n\nif __name__ == "__main__":\n    import uvicorn\n\n    uvicorn.run(app, interface="asgi3", port=4190)\n```\n',
    'author': 'abersheeran',
    'author_email': 'me@abersheeran.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abersheeran/index.py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import re
from dataclasses import dataclass, field
from typing import Any, Dict, Generator, List, Optional, Pattern, Tuple, Union

from ..types import ASGIApp
from .convertors import Convertor, PathConvertor, compile_path


@dataclass
class TreeNode:
    characters: str
    re_pattern: Optional[Pattern] = None
    param_convertors: Dict[str, Convertor] = field(default_factory=dict)
    next_nodes: List["TreeNode"] = field(default_factory=list)
    endpoint: Optional[ASGIApp] = None


def find_common_prefix(x: str, y: str) -> str:
    """
    find the longest common prefix of x and y
    """
    for i in range(min(len(x), len(y))):
        if x[i] != y[i]:
            return x[:i]
    return x[: i + 1]


def append(
    point: TreeNode, path_format: str, param_convertors: Dict[str, Convertor]
) -> TreeNode:
    """
    Construct the node corresponding to the specified path and return.

    The order of child nodes under the same node is determined by the order of addition.
    """
    if not path_format:
        return point

    if path_format[0] == "{":
        length = path_format.find("}") + 1
        param_name = path_format[1 : length - 1]
        convertor = param_convertors[param_name]
        re_pattern = re.compile(convertor.regex)
        if isinstance(convertor, PathConvertor) and path_format[-1] != "}":
            raise ValueError(
                "`PathConvertor` is only allowed to appear at the end of path"
            )
        for node in filter(lambda node: node.re_pattern is not None, point.next_nodes):
            if (node.re_pattern == re_pattern) != (node.characters == param_name):
                raise ValueError(
                    "The same regular matching is used in the same position"
                    + ", but the parameter names are different."
                )
            if node.characters == param_name:
                return append(node, path_format[length:], param_convertors)

        new_node = TreeNode(characters=param_name, re_pattern=re_pattern)
        point.next_nodes.insert(0, new_node)
        return append(new_node, path_format[length:], param_convertors)

    length = path_format.find("{")
    if length == -1:
        length = len(path_format)

    for node in filter(lambda node: node.re_pattern is None, point.next_nodes):
        prefix = find_common_prefix(node.characters, path_format[:length])
        if prefix == "":
            continue
        if node.characters == prefix:
            return append(node, path_format[len(prefix) :], param_convertors)

        node_index = point.next_nodes.index(node)
        prefix_node = TreeNode(characters=prefix)
        point.next_nodes[node_index] = prefix_node
        node.characters = node.characters[len(prefix) :]
        prefix_node.next_nodes.insert(0, node)
        if path_format[:length] == prefix:
            return append(prefix_node, path_format[length:], param_convertors)

        new_node = TreeNode(characters=path_format[len(prefix) : length])
        prefix_node.next_nodes.insert(0, new_node)
        return append(new_node, path_format[length:], param_convertors)

    new_node = TreeNode(characters=path_format[:length])
    point.next_nodes.insert(0, new_node)
    return append(new_node, path_format[length:], param_convertors)


def search(
    point: TreeNode, path: str, params: Dict[str, str]
) -> Optional[Tuple[Dict[str, str], TreeNode]]:
    """
    Find a suitable route
    """
    stack: List[Tuple[str, TreeNode]] = [(path, point)]

    while stack:
        path, point = stack.pop(len(stack) - 1)

        if point.re_pattern is None:
            length = len(point.characters)
            if path[:length] != point.characters:
                continue
        else:
            none_or_match = re.match(point.re_pattern, path)
            if none_or_match is None:
                continue
            result = none_or_match.group()
            params[point.characters] = result
            length = len(result)

        path = path[length:]
        if not path:  # path == "", found the first suitable route
            [
                params.pop(name)
                for name in set(params.keys()) - set(point.param_convertors.keys())
            ]
            return params, point

        for node in point.next_nodes:
            stack.append((path, node))

    return None


class RadixTree:
    def __init__(self) -> None:
        self.root = TreeNode("/")

    def append(self, path: str, endpoint: ASGIApp) -> None:
        if path[0] != "/":
            raise ValueError('path must start with "/"')
        path_format, param_convertors = compile_path(path)
        point = append(self.root, path_format[1:], param_convertors)

        if point.endpoint is not None:
            raise ValueError(f"Routing conflict: {path}")

        point.endpoint = endpoint
        point.param_convertors = param_convertors

    def search(
        self, path: str
    ) -> Union[Tuple[Dict[str, Any], ASGIApp], Tuple[None, None]]:
        result = search(self.root, path, {})
        if result is None:
            return None, None

        raw_params, node = result
        if node.endpoint is None:
            return None, None

        return (
            {
                name: node.param_convertors[name].convert(value)
                for name, value in raw_params.items()
            },
            node.endpoint,
        )

    def iterator(self) -> Generator[Tuple[str, ASGIApp], None, None]:
        stack: List[Tuple[str, TreeNode]] = [(self.root.characters, self.root)]

        while stack:
            characters, point = stack.pop(len(stack) - 1)
            for node in point.next_nodes:
                if node.re_pattern is not None:
                    stack.append((f"{characters}{{{node.characters}}}", node))
                else:
                    stack.append((f"{characters}{node.characters}", node))
            if point.endpoint:
                yield characters, point.endpoint



import functools
import logging
import typing
from inspect import isclass, signature

from pydantic import BaseModel, ValidationError

from ..concurrency import keepasync
from .request import Request
from .responses import Response

HTTP_METHOD_NAMES = [
    "get",
    "post",
    "put",
    "patch",
    "delete",
    "head",
    "options",
    "trace",
]

logger = logging.getLogger(__name__)


class ParamsValidationError(Exception):
    def __init__(self, validation_error: ValidationError) -> None:
        self.ve = validation_error


def parse_params(function: typing.Callable) -> typing.Callable:
    """
    parse function params "path", "query", "header", "cookie", "body"
    """
    param_names = ("path", "query", "header", "cookie", "body")
    sig = signature(function)

    incorrect_keys = [
        param_name
        for param_name in param_names
        if (
            param_name in sig.parameters
            and not issubclass(sig.parameters[param_name].annotation, BaseModel)
        )
    ]
    if incorrect_keys:
        raise TypeError(
            f"Params {incorrect_keys} annotation is incorrect in `{function.__name__}`. "
            + "You should inherit `pydantic.BaseModel`."
        )

    __params__ = {
        param_name: sig.parameters[param_name].annotation
        for param_name in param_names
        if (
            param_name in sig.parameters
            and issubclass(sig.parameters[param_name].annotation, BaseModel)
        )
    }
    if __params__:
        setattr(function, "__params__", __params__)
    return function


def merge_list(
    raw: typing.List[typing.Tuple[str, str]]
) -> typing.Dict[str, typing.Union[typing.List[str], str]]:
    """
    If there are values with the same key value, they are merged into a List.
    """
    d: typing.Dict[str, typing.Union[typing.List[str], str]] = {}
    for k, v in raw:
        if k in d:
            if isinstance(d[k], list):
                typing.cast(typing.List, d[k]).append(v)
            else:
                d[k] = [typing.cast(str, d[k]), v]
        else:
            d[k] = v
    return d


async def bound_params(handler: typing.Callable, request: Request) -> typing.Callable:
    """
    bound parameters "path", "query", "header", "cookie", "body" to the view function
    """
    if isclass(handler):
        return handler

    __params__ = getattr(handler, "__params__", None)
    if not __params__:
        return handler

    params: typing.Dict[str, BaseModel] = {}

    try:
        # try to get parameters model and parse
        if "path" in __params__:
            params["path"] = __params__["path"](**request.path_params)

        if "query" in __params__:
            params["query"] = __params__["query"](
                **merge_list(request.query_params.multi_items())
            )

        if "header" in __params__:
            params["header"] = __params__["header"](
                **merge_list(request.headers.items())
            )

        if "cookie" in __params__:
            params["cookie"] = __params__["cookie"](**request.cookies)

        # try to get body model and parse
        if "body" in __params__:
            if request.headers.get("Content-Type") == "application/json":
                _body_data = await request.json()
            else:
                _body_data = await request.form()
            params["body"] = __params__["body"](**_body_data)

    except ValidationError as e:
        raise ParamsValidationError(e)

    return functools.partial(handler, **params)


class ViewMeta(keepasync(*HTTP_METHOD_NAMES)):  # type: ignore
    def __init__(
        cls,
        name: str,
        bases: typing.Tuple[type],
        namespace: typing.Dict[str, typing.Any],
    ):
        for function_name in filter(
            lambda key: key in HTTP_METHOD_NAMES, namespace.keys()
        ):
            function = namespace[function_name]
            namespace[function_name] = parse_params(function)

        setattr(
            cls,
            "__methods__",
            [m.upper() for m in HTTP_METHOD_NAMES if hasattr(cls, m)],
        )

        super().__init__(name, bases, namespace)


class HTTPView(metaclass=ViewMeta):  # type: ignore
    if typing.TYPE_CHECKING:
        __methods__: typing.List[str]

    def __init__(self, request: Request) -> None:
        self.request = request

    def __await__(self) -> typing.Generator[typing.Any, None, Response]:
        return self.__impl__().__await__()

    async def __impl__(self) -> typing.Any:
        # Try to dispatch to the right method; if a method doesn't exist,
        # defer to the error handler.
        handler = getattr(
            self, self.request.method.lower(), self.http_method_not_allowed
        )

        handler = await bound_params(handler, self.request)

        return await handler()

    async def http_method_not_allowed(self) -> Response:
        if self.request.method not in ("GET", "HEAD"):
            status_code = 405
        else:
            status_code = 404

        return Response(
            status_code=status_code,
            headers={"Allow": ", ".join(self.__methods__), "Content-Length": "0"},
        )

    async def options(self) -> Response:
        """Handle responding to requests for the OPTIONS HTTP verb."""
        return Response(
            headers={"Allow": ", ".join(self.__methods__), "Content-Length": "0"}
        )


def only_allow(method: str = "", func: typing.Callable = None) -> typing.Callable:
    """
    Only allow the function to accept one type of request

    example:
        @only_allow("get")
        async def handle(request):
            ...
    or
        handle = only_allow("get", handle)
    """

    if method not in HTTP_METHOD_NAMES:
        raise ValueError(f"method must be in {HTTP_METHOD_NAMES}")

    if func is None:
        return lambda func: only_allow(method, func)

    func = parse_params(func)
    setattr(func, "__method__", method.upper())

    @functools.wraps(func)
    async def wrapper(*args, **kwargs) -> typing.Any:
        request = args[0]
        if request.method.lower() == method.lower():
            handler = await bound_params(func, request)  # type: ignore
            return await handler(*args, **kwargs)

        if request.method == "OPTIONS":
            return Response(headers={"Allow": method, "Content-Length": "0"})
        elif request.method in ("GET", "HEAD"):
            status_code = 404
        else:
            status_code = 405
        return Response(status_code=status_code)

    return wrapper
